library ABuffSpell requires Table, ABuff, SpellEvent, optional GroupUtils

//*****************************************************************
//*  ABUFF SPELL
//*
//*  written by: Anitarf
//*  requires: -Table
//*            -ABuff
//*            -SpellEvent
//*  optional: -GroupUtils
//*
//*  This library automates the use of aBuffs as effect of spells.
//*  Once an aBuffType is linked to an ability using the 
//*  NewABuffTargetSpell or NewABuffAreaSpell functions, that
//*  ability will automatically apply buffs of that type when
//*  cast. A targeted ABuff spell can either target a unit or
//*  have no target, in which case it will be assumed it targets
//*  the caster. An area spell can have any kind of targeting,
//*  unit, point or none. Properties of the spell (the duration
//*  of the buff, the radius of the area of effect and allowed
//*  targets) are specified through functions that follow the
//*  function interfaces specified below.
//*
//*  In order for aBuffs to be properly applied on all units that
//*  are highlighted by the WC3's area-targeting cursor, the
//*  GroupUtils library is required. The requirement is only an
//*  optional one because non-area spells work fine without it,
//*  if a map contains no area-target spells then there is no
//*  need to import GroupUtils.
//*
//*  To define an aBuff spell, use these two functions:
//*
//*    function NewABuffTargetSpell takes integer abilityId, aBuffType buffType, SpellPropertyFunction buffDuration returns nothing
//*    function NewABuffAreaSpell takes integer abilityId, aBuffType buffType, SpellPropertyFunction buffDuration, SpellPropertyFunction abilityAoE, TargetFilterFunction allowedTargets returns nothing
//*
//*  The functions only accept buff types in the STANDARD and
//*  STACKING categories.
//*****************************************************************

    function interface SpellPropertyFunction takes integer level returns real
    function interface TargetFilterFunction takes unit caster, unit target returns boolean

// ================================================================

    private struct spell
        integer abilityId
        aBuffType buffType
        
        SpellPropertyFunction buffDuration
        SpellPropertyFunction abilityAoE
        TargetFilterFunction allowedTargets
        
        private static Table table
        static method create takes integer abilityId, aBuffType buffType, SpellPropertyFunction buffDuration, SpellPropertyFunction abilityAoE, TargetFilterFunction allowedTargets returns spell
            local spell this=spell.allocate()
            set .abilityId=abilityId
            set .buffType=buffType
            set .buffDuration=buffDuration
            set .abilityAoE=abilityAoE
            set .allowedTargets=allowedTargets
            debug if spell.table[abilityId]!=0 then
            debug    call BJDebugMsg("NewABuffSpell error: ability rawcode already associated with an ABuffSpell!") 
            debug else
            set spell.table[abilityId]=this
            debug endif
            return this
        endmethod
        static method get takes integer abilityId returns spell
            return spell.table[abilityId]
        endmethod
        
// ================================================================

        method applyBuff takes unit t, unit c, integer l returns nothing
            if .buffDuration==0 then
                call ABuffApply(.buffType, t, c, 0.0, l, 0)
            else
                call ABuffApply(.buffType, t, c, .buffDuration.evaluate(l), l, 0)
            endif
        endmethod

        static group g = CreateGroup()
        static spell current = 0
        static unit caster = null
        static integer level = 0
        static method unitEnum takes nothing returns boolean
            if IsUnitType(GetFilterUnit(), UNIT_TYPE_DEAD) then
            elseif current.allowedTargets==0 then
                call .current.applyBuff(GetFilterUnit(), .caster, .level)
            elseif current.allowedTargets.evaluate(.caster, GetFilterUnit()) then
                call .current.applyBuff(GetFilterUnit(), .caster, .level)
            endif
            return false
        endmethod

        private static method spellCast takes nothing returns nothing
            local spell this = spell.get(SpellEvent.AbilityId)
            local real x = SpellEvent.TargetX
            local real y = SpellEvent.TargetY
            local unit t = SpellEvent.TargetUnit

            local spell s = .current // remember previous values
            local unit c = .caster // in case this is interrupting
            local integer l = .level // another spellCast execution.
            set .current = this
            set .caster = SpellEvent.CastingUnit
            set .level = GetUnitAbilityLevel(.caster, .abilityId)

            if this!=0 then
                if t==null and x==0.0 and y==0.0 then
                    // no-target spell, make the caster the target
                    set t=.caster
                    set x=GetUnitX(t)
                    set y=GetUnitY(t)
                endif

                if .abilityAoE==0 then
                    // no AoE function, apply the buff only on the main target
                    if t==null then
                        debug call BJDebugMsg("ABuffSpell error: no target found for a single-target spell.")
                    else
                        call .applyBuff(t, .caster, .level)
                    endif
                else
                    // apply the buff on all the valid units in the area of effect
                    static if LIBRARY_GroupUtils then
                    call GroupEnumUnitsInArea(.g, x,y,.abilityAoE.evaluate(.level), Condition(function spell.unitEnum))
                    else
                    call GroupEnumUnitsInRange(.g, x,y,.abilityAoE.evaluate(.level), Condition(function spell.unitEnum))
                    endif
                endif
            endif

            set .current = s
            set .caster = c
            set .level = l
            set c=null
        endmethod

        private static method onInit takes nothing returns nothing
            set spell.table=Table.create()
            call RegisterSpellEffectResponse(0, spell.spellCast)
        endmethod
    endstruct

// ================================================================
//  MAIN USER FUNCTIONS

    function NewABuffAreaSpell takes integer abilityId, aBuffType buffType, SpellPropertyFunction buffDuration, SpellPropertyFunction abilityAoE, TargetFilterFunction allowedTargets returns nothing
        if buffType.category==ABuff_STANDARD or buffType.category==ABuff_STACKING then
            call spell.create(abilityId, buffType, buffDuration, abilityAoE, allowedTargets)
        else
            debug call BJDebugMsg("NewABuffSpell error: the function only accepts buff types in the STANDARD and STACKING categories.")
        endif
    endfunction

    function NewABuffTargetSpell takes integer abilityId, aBuffType buffType, SpellPropertyFunction buffDuration returns nothing
        call NewABuffAreaSpell(abilityId, buffType, buffDuration, 0,0)
    endfunction
    
endlibrary